#ifndef _math_evaluator_h_
#define _math_evaluator_h_

#define OP_CODE_PUSH						100
#define OP_CODE_PUSH_VARIABLE				101
#define OP_CODE_PUSH_VARIABLE_REFERENCE		102

#define OP_CODE_ADD					201
#define OP_CODE_SUBTRACT			202
#define OP_CODE_MULTIPLY			203
#define OP_CODE_DIVIDE				204
#define OP_CODE_NEGATE				205
#define OP_CODE_MODULUS				206
#define OP_CODE_EXP					207
#define OP_CODE_MIN					208
#define OP_CODE_MAX					209
#define OP_CODE_ABSMIN				210
#define OP_CODE_ABSMAX				211

#define OP_CODE_SIN					300
#define OP_CODE_COS					301
#define OP_CODE_TAN					302
#define OP_CODE_ABS					303
#define OP_CODE_ATAN				304
#define OP_CODE_LOG					305
#define OP_CODE_LOG10				306
#define OP_CODE_ABSSIN				307
#define OP_CODE_ABSCOS				308
#define OP_CODE_RAND				309
#define OP_CODE_RANDN				310
#define OP_CODE_IF					311
#define OP_CODE_SIGN				312
#define OP_CODE_MIN_FUNC			313
#define OP_CODE_MAX_FUNC			314
#define OP_CODE_EQUAL				315
#define OP_CODE_BELOW				316
#define OP_CODE_ABOVE				317
#define OP_CODE_POW					318

#define OP_CODE_ASSIGN				400
#define	OP_CODE_CLEAR				401

#define OP							1
#define CONSTANT					2
#define VAR							3
#define FUNCTION					4
#define COMMA						5

typedef struct token_s 
{
	int type;
	int val;
	double fval;
	char name[256];
	struct token_s* next;
} token;

struct OP_CODE 
{
	unsigned int	nOpCode;
	float			fValue;
	unsigned int	nParam1;
};

/***********************************************************************************
CVariable - the atomic information about a variable
***********************************************************************************/
class CVariable
{
public:

	float fValue;
	CString strName;
};

/***********************************************************************************
CVariableInfo - wraps the variable and stores ownership information
***********************************************************************************/
class CVariableInfo
{
public:

	CVariableInfo(const char * pName, float fValue)
	{
		m_pVariable = new CVariable;
		m_pVariable->fValue = fValue;
		m_pVariable->strName = pName;
		m_bOwned = TRUE;
	}

	CVariableInfo(CVariable * pVariable)
	{
		m_pVariable = pVariable;
		m_bOwned = FALSE;
	}

	~CVariableInfo()
	{
		if (m_bOwned && m_pVariable)
		{
			delete m_pVariable;
			m_pVariable = NULL;
		}
	}

	CVariable * m_pVariable;
	BOOL m_bOwned;
};	

/***********************************************************************************
CVariableSpace - the container for variables (can share, merge, etc.)
***********************************************************************************/
class CVariableSpace
{
public:

	CVariableSpace()
	{
	}

	~CVariableSpace()
	{
		Clear();
	}

	// functions
	void SetVariableValue(long nIndex, float fValue);
	void SetVariableValue(const char * pVariableName, float fValue);
	long GetVariableIndex(const char * pVariableName, BOOL bCreate = TRUE);
	float GetVariableValue(const char * pVariableName);
	float GetVariableValue(long nIndex);
	void AppendVariables(CVariableSpace * pVariableSpace, BOOL bClearExisting = FALSE);
	void Clear();

	// data
	CArray<CVariableInfo *, CVariableInfo *> m_aryVariables;
};

/***********************************************************************************
CMathParser - parse out the math equation
***********************************************************************************/
class CMathParser
{
public:

	long LoadOpCodes(char * pExpression, CArray<OP_CODE, OP_CODE &> * paryOpCodes, CVariableSpace * pVariables);

protected:

	long				AddOPCode(unsigned int nOpCode, float fValue, unsigned int nParam1 = 0);

	token*				tokenize(char* in);
	token*				parse_base(token* in);
	token*				parse_assign(token* in);
	token*				parse_const(token* in);
	token*				parse_paren(token* in);
	token*				parse_mult(token* in);
	token*				parse_plus(token* in);
	token*				parse_exp(token* in);
	token*				parse_minmax(token* in);
	token*				parse_var_ref(token* in);
	void				free_token(token* in);
	void				parse(token* in);

	bool				m_bAssign;

	CArray<OP_CODE, OP_CODE &> * m_paryOpCodes;
	CVariableSpace * m_pVariables;
};

/***********************************************************************************
CMathEvaluator - the evaluation class
***********************************************************************************/
class CMathEvaluator
{
public:
	CMathEvaluator();
	~CMathEvaluator();
	
	void SetEquation(const char * pEquation);
	float Evaluate();
	
	void SetVariableSpace(CVariableSpace * pVariables);

	CVariableSpace * m_pVariables;

private:

	CArray<OP_CODE, OP_CODE &>	m_aryOpCodes;
	
	float				m_fStack[512];
	long				m_nStackIndex;
	BOOL				m_bIsValid;

	float				m_arySinTable[1024];
	float				m_aryCosTable[1024];

	BOOL				m_bOwnsVariables;

	int Optimize();
};

#endif //_math_evaluator_h_